<span class="clarification"
      tabindex="0"
      role="note"
      aria-label="Clarification">
    <span data-tooltip="{{ clarification }}"
          data-tooltip-show-on-click="true"
          class="fa fa-info-circle"
          aria-hidden="true"></span>
    <span class="sr">({{ clarification }})</span>
</span>
