{% if status_id == '' %}
    <span class="status {% if hide_correctness == True %}{% else %}{{ status.classname }}{% endif %}"
          data-tooltip="{% if hide_correctness == True %}{% else %}{{ status.display_tooltip }}{% endif %}">
    {% else %}
        <span class="status {% if hide_correctness == True %}{% else %}{{ status.classname }}{% endif %}"
              id="status_{{ status_id }}"
              data-tooltip="{% if hide_correctness == True %}{% else %}{{ status.display_tooltip }}{% endif %}">
        {% endif %}
        {% if not hide_correctness %}
            <span class="sr">{{ status.display_name }}</span><span class="status-icon" aria-hidden="true"></span>
        {% endif %}
    </span>
