{% load i18n %}
<div class="notification {{ notification_type }} notification-{{ notification_name }} {% if is_hidden != False %}is-hidden{% endif %}"
     tabindex="-1">
    <span class="icon fa {{ notification_icon }}" aria-hidden="true"></span>
    <span class="notification-message"
          aria-describedby="{{ short_id }}-problem-title">{{ notification_message }}</span>
    <div class="notification-btn-wrapper">
        {% if notification_name == 'hint' %}
            <button type="button"
                    class="btn btn-default btn-small notification-btn hint-button">
                {% trans "Next Hint" as next_hint_text %}{{ next_hint_text|force_escape }}
            </button>
        {% endif %}
        <button type="button"
                class="btn btn-default btn-small notification-btn review-btn sr">
            {% trans "Review" as review_text %}{{ review_text|force_escape }}
        </button>
    </div>
</div>
